#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from glanceclient import client as glance_client
from neutronclient.neutron import client as neutron_client
from novaclient import api_versions
from novaclient import client as nova_client
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from octavia.common import keystone

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

GLANCE_VERSION = '2'
NEUTRON_VERSION = '2.0'
NOVA_VERSION = '2.15'


class NovaAuth(object):
    nova_client = None

    @classmethod
    def get_nova_client(cls, region, service_name=None, endpoint=None,
                        endpoint_type='publicURL', insecure=False,
                        cacert=None):
        """Create nova client object.

        :param region: The region of the service
        :param service_name: The name of the nova service in the catalog
        :param endpoint: The endpoint of the service
        :param endpoint_type: The type of the endpoint
        :param insecure: Turn off certificate validation
        :param cacert: CA Cert file path
        :return: a Nova Client object.
        :raises Exception: if the client cannot be created
        """
        ksession = keystone.KeystoneSession()
        if not cls.nova_client:
            kwargs = {'region_name': region,
                      'session': ksession.get_session(),
                      'endpoint_type': endpoint_type,
                      'insecure': insecure}
            if service_name:
                kwargs['service_name'] = service_name
            if endpoint:
                kwargs['endpoint_override'] = endpoint
            if cacert:
                kwargs['cacert'] = cacert
            try:
                cls.nova_client = nova_client.Client(
                    version=api_versions.APIVersion(NOVA_VERSION), **kwargs)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Error creating Nova client.")
        return cls.nova_client


class NeutronAuth(object):
    neutron_client = None

    @classmethod
    def get_neutron_client(cls, region, service_name=None, endpoint=None,
                           endpoint_type='publicURL', insecure=False,
                           ca_cert=None):
        """Create neutron client object.

        :param region: The region of the service
        :param service_name: The name of the neutron service in the catalog
        :param endpoint: The endpoint of the service
        :param endpoint_type: The endpoint_type of the service
        :param insecure: Turn off certificate validation
        :param ca_cert: CA Cert file path
        :return: a Neutron Client object.
        :raises Exception: if the client cannot be created
        """
        ksession = keystone.KeystoneSession()
        if not cls.neutron_client:
            kwargs = {'region_name': region,
                      'session': ksession.get_session(),
                      'endpoint_type': endpoint_type,
                      'insecure': insecure}
            if service_name:
                kwargs['service_name'] = service_name
            if endpoint:
                kwargs['endpoint_override'] = endpoint
            if ca_cert:
                kwargs['ca_cert'] = ca_cert
            try:
                cls.neutron_client = neutron_client.Client(
                    NEUTRON_VERSION, **kwargs)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Error creating Neutron client.")
        return cls.neutron_client


class GlanceAuth(object):
    glance_client = None

    @classmethod
    def get_glance_client(cls, region, service_name=None, endpoint=None,
                          endpoint_type='publicURL', insecure=False,
                          cacert=None):
        """Create glance client object.

        :param region: The region of the service
        :param service_name: The name of the glance service in the catalog
        :param endpoint: The endpoint of the service
        :param endpoint_type: The endpoint_type of the service
        :param insecure: Turn off certificate validation
        :param cacert: CA Cert file path
        :return: a Glance Client object.
        :raises Exception: if the client cannot be created
        """
        ksession = keystone.KeystoneSession()
        if not cls.glance_client:
            kwargs = {'region_name': region,
                      'session': ksession.get_session(),
                      'interface': endpoint_type}
            if service_name:
                kwargs['service_name'] = service_name
            if endpoint:
                kwargs['endpoint'] = endpoint
                if endpoint.startswith("https"):
                    kwargs['insecure'] = insecure
                    kwargs['cacert'] = cacert
            try:
                cls.glance_client = glance_client.Client(
                    GLANCE_VERSION, **kwargs)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Error creating Glance client.")
        return cls.glance_client



# Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from stevedore import driver as stevedore_driver

from octavia.common import constants

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class Endpoint(object):

    # API version history:
    #   1.0 - Initial version.
    target = messaging.Target(
        namespace=constants.RPC_NAMESPACE_CONTROLLER_AGENT,
        version='1.1')

    def __init__(self):
        self.worker = stevedore_driver.DriverManager(
            namespace='octavia.plugins',
            name=CONF.octavia_plugins,
            invoke_on_load=True
        ).driver

    def create_load_balancer(self, context, load_balancer_id,
                             flavor=None):
        LOG.info('Creating load balancer \'%s\'...', load_balancer_id)
        self.worker.create_load_balancer(load_balancer_id, flavor)

    def update_load_balancer(self, context, load_balancer_id,
                             load_balancer_updates):
        LOG.info('Updating load balancer \'%s\'...', load_balancer_id)
        self.worker.update_load_balancer(load_balancer_id,
                                         load_balancer_updates)

    def delete_load_balancer(self, context, load_balancer_id, cascade=False):
        LOG.info('Deleting load balancer \'%s\'...', load_balancer_id)
        self.worker.delete_load_balancer(load_balancer_id, cascade)

    def failover_load_balancer(self, context, load_balancer_id):
        LOG.info('Failing over amphora in load balancer \'%s\'...',
                 load_balancer_id)
        self.worker.failover_loadbalancer(load_balancer_id)

    def failover_amphora(self, context, amphora_id):
        LOG.info('Failing over amphora \'%s\'...',
                 amphora_id)
        self.worker.failover_amphora(amphora_id)

    def create_listener(self, context, listener_id):
        LOG.info('Creating listener \'%s\'...', listener_id)
        self.worker.create_listener(listener_id)

    def update_listener(self, context, listener_id, listener_updates):
        LOG.info('Updating listener \'%s\'...', listener_id)
        self.worker.update_listener(listener_id, listener_updates)

    def delete_listener(self, context, listener_id):
        LOG.info('Deleting listener \'%s\'...', listener_id)
        self.worker.delete_listener(listener_id)

    def create_pool(self, context, pool_id):
        LOG.info('Creating pool \'%s\'...', pool_id)
        self.worker.create_pool(pool_id)

    def update_pool(self, context, pool_id, pool_updates):
        LOG.info('Updating pool \'%s\'...', pool_id)
        self.worker.update_pool(pool_id, pool_updates)

    def delete_pool(self, context, pool_id):
        LOG.info('Deleting pool \'%s\'...', pool_id)
        self.worker.delete_pool(pool_id)

    def create_health_monitor(self, context, health_monitor_id):
        LOG.info('Creating health monitor \'%s\'...', health_monitor_id)
        self.worker.create_health_monitor(health_monitor_id)

    def update_health_monitor(self, context, health_monitor_id,
                              health_monitor_updates):
        LOG.info('Updating health monitor \'%s\'...', health_monitor_id)
        self.worker.update_health_monitor(health_monitor_id,
                                          health_monitor_updates)

    def delete_health_monitor(self, context, health_monitor_id):
        LOG.info('Deleting health monitor \'%s\'...', health_monitor_id)
        self.worker.delete_health_monitor(health_monitor_id)

    def create_member(self, context, member_id):
        LOG.info('Creating member \'%s\'...', member_id)
        self.worker.create_member(member_id)

    def update_member(self, context, member_id, member_updates):
        LOG.info('Updating member \'%s\'...', member_id)
        self.worker.update_member(member_id, member_updates)

    def batch_update_members(self, context, old_member_ids, new_member_ids,
                             updated_members):
        updated_member_ids = [m.get('id') for m in updated_members]
        LOG.info(
            'Batch updating members: old=\'%(old)s\', new=\'%(new)s\', '
            'updated=\'%(updated)s\'...',
            {'old': old_member_ids, 'new': new_member_ids,
             'updated': updated_member_ids})
        self.worker.batch_update_members(
            old_member_ids, new_member_ids, updated_members)

    def delete_member(self, context, member_id):
        LOG.info('Deleting member \'%s\'...', member_id)
        self.worker.delete_member(member_id)

    def create_l7policy(self, context, l7policy_id):
        LOG.info('Creating l7policy \'%s\'...', l7policy_id)
        self.worker.create_l7policy(l7policy_id)

    def update_l7policy(self, context, l7policy_id, l7policy_updates):
        LOG.info('Updating l7policy \'%s\'...', l7policy_id)
        self.worker.update_l7policy(l7policy_id, l7policy_updates)

    def delete_l7policy(self, context, l7policy_id):
        LOG.info('Deleting l7policy \'%s\'...', l7policy_id)
        self.worker.delete_l7policy(l7policy_id)

    def create_l7rule(self, context, l7rule_id):
        LOG.info('Creating l7rule \'%s\'...', l7rule_id)
        self.worker.create_l7rule(l7rule_id)

    def update_l7rule(self, context, l7rule_id, l7rule_updates):
        LOG.info('Updating l7rule \'%s\'...', l7rule_id)
        self.worker.update_l7rule(l7rule_id, l7rule_updates)

    def delete_l7rule(self, context, l7rule_id):
        LOG.info('Deleting l7rule \'%s\'...', l7rule_id)
        self.worker.delete_l7rule(l7rule_id)

    def update_amphora_agent_config(self, context, amphora_id):
        LOG.info('Updating amphora \'%s\' agent configuration...',
                 amphora_id)
        self.worker.update_amphora_agent_config(amphora_id)



# Copyright 2015 Hewlett Packard Enterprise Development Company LP
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Keepalived configuration datamodel

Revision ID: 1e4c1d83044c
Revises: 5a3ee5472c31
Create Date: 2015-08-06 10:39:54.998797

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '1e4c1d83044c'
down_revision = '5a3ee5472c31'


def upgrade():
    op.create_table(
        u'vrrp_auth_method',
        sa.Column(u'name', sa.String(36), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'vrrp_auth_method',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'PASS'},
            {'name': 'AH'}
        ]
    )

    op.create_table(
        u'vrrp_group',
        sa.Column(u'load_balancer_id', sa.String(36), nullable=False),
        sa.Column(u'vrrp_group_name', sa.String(36), nullable=True),
        sa.Column(u'vrrp_auth_type', sa.String(16), nullable=True),
        sa.Column(u'vrrp_auth_pass', sa.String(36), nullable=True),
        sa.Column(u'advert_int', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint(u'load_balancer_id'),
        sa.ForeignKeyConstraint([u'load_balancer_id'], [u'load_balancer.id'],
                                name=u'fk_vrrp_group_load_balancer_id'),
        sa.ForeignKeyConstraint([u'vrrp_auth_type'],
                                [u'vrrp_auth_method.name'],
                                name=u'fk_load_balancer_vrrp_auth_method_name')
    )

    op.add_column(
        u'listener',
        sa.Column(u'peer_port', sa.Integer(), nullable=True)
    )

    op.add_column(
        u'amphora',
        sa.Column(u'vrrp_interface', sa.String(16), nullable=True)
    )

    op.add_column(
        u'amphora',
        sa.Column(u'vrrp_id', sa.Integer(), nullable=True)
    )

    op.add_column(
        u'amphora',
        sa.Column(u'vrrp_priority', sa.Integer(), nullable=True)
    )



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Add a column busy in table amphora health

Revision ID: 543f5d8e4e56
Revises: 2351ea316465
Create Date: 2015-07-27 11:32:16.685383

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '543f5d8e4e56'
down_revision = '2351ea316465'


def upgrade():
    op.add_column(u'amphora_health',
                  sa.Column(u'busy', sa.Boolean(), nullable=False))



#    Copyright 2017 GoDaddy
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add timestamps and operating_status to healthmonitor

Revision ID: 52377704420e
Revises: d85ca7258d21
Create Date: 2017-04-13 08:58:18.078170

"""

from alembic import op
import sqlalchemy as sa

from octavia.common import constants

# revision identifiers, used by Alembic.
revision = '52377704420e'
down_revision = 'd85ca7258d21'


def upgrade():
    op.add_column(
        u'health_monitor',
        sa.Column(u'created_at', sa.DateTime(), nullable=True)
    )
    op.add_column(
        u'health_monitor',
        sa.Column(u'updated_at', sa.DateTime(), nullable=True)
    )

    op.add_column(u'health_monitor',
                  sa.Column(u'operating_status',
                            sa.String(16),
                            nullable=False,
                            server_default=constants.ONLINE)
                  )
    op.alter_column(u'health_monitor', u'operating_status',
                    existing_type=sa.String(16), server_default=None)

    op.create_foreign_key(
        u'fk_health_monitor_operating_status_name', u'health_monitor',
        u'operating_status', [u'operating_status'], [u'name']
    )



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from octavia.common import constants
from octavia.tests.functional.api.v1 import base


class TestPool(base.BaseAPITest):

    def setUp(self):
        super(TestPool, self).setUp()
        self.lb = self.create_load_balancer(
            {'subnet_id': uuidutils.generate_uuid()})
        self.lb = self.set_lb_status(self.lb.get('id'))
        self.listener = self.create_listener(self.lb.get('id'),
                                             constants.PROTOCOL_HTTP, 80)
        self.lb = self.set_lb_status(self.lb.get('id'))
        self.listener = self.get_listener(self.lb.get('id'),
                                          self.listener.get('id'))
        self.pools_path = self.POOLS_PATH.format(lb_id=self.lb.get('id'))
        self.pool_path = self.pools_path + '/{pool_id}'
        self.pools_path_with_listener = (self.pools_path +
                                         '?listener_id={listener_id}')
        self.pools_path_deprecated = self.DEPRECATED_POOLS_PATH.format(
            lb_id=self.lb.get('id'), listener_id=self.listener.get('id'))
        self.pool_path_deprecated = self.pools_path_deprecated + '/{pool_id}'

    def test_get(self):
        api_pool = self.create_pool_sans_listener(
            self.lb.get('id'), constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        response = self.get(self.pool_path.format(pool_id=api_pool.get('id')))
        response_body = response.json
        self.assertEqual(api_pool, response_body)

    def test_bad_get(self):
        self.get(self.pool_path.format(pool_id=uuidutils.generate_uuid()),
                 status=404)

    def test_get_all(self):
        api_pool = self.create_pool_sans_listener(
            self.lb.get('id'), constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        response = self.get(self.pools_path)
        response_body = response.json
        self.assertIsInstance(response_body, list)
        self.assertEqual(1, len(response_body))
        self.assertEqual(api_pool.get('id'), response_body[0].get('id'))

    def test_get_all_with_listener(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        response = self.get(self.pools_path_with_listener.format(
            listener_id=self.listener.get('id')))
        response_body = response.json
        self.assertIsInstance(response_body, list)
        self.assertEqual(1, len(response_body))
        self.assertEqual(api_pool.get('id'), response_body[0].get('id'))

    def test_get_all_with_bad_listener(self):
        self.get(self.pools_path_with_listener.format(
            listener_id='bad_id'), status=404, expect_errors=True)

    def test_empty_get_all(self):
        response = self.get(self.pools_path)
        response_body = response.json
        self.assertIsInstance(response_body, list)
        self.assertEqual(0, len(response_body))

    def test_create(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assertEqual(constants.PROTOCOL_HTTP, api_pool.get('protocol'))
        self.assertEqual(constants.LB_ALGORITHM_ROUND_ROBIN,
                         api_pool.get('lb_algorithm'))
        self.assertIsNotNone(api_pool.get('created_at'))
        self.assertIsNone(api_pool.get('updated_at'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_with_proxy_protocol(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_PROXY,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assertEqual(constants.PROTOCOL_PROXY, api_pool.get('protocol'))
        self.assertEqual(constants.LB_ALGORITHM_ROUND_ROBIN,
                         api_pool.get('lb_algorithm'))
        self.assertIsNotNone(api_pool.get('created_at'))
        self.assertIsNone(api_pool.get('updated_at'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_sans_listener(self):
        api_pool = self.create_pool_sans_listener(
            self.lb.get('id'), constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN)
        self.assertEqual(constants.PROTOCOL_HTTP, api_pool.get('protocol'))
        self.assertEqual(constants.LB_ALGORITHM_ROUND_ROBIN,
                         api_pool.get('lb_algorithm'))
        # Make sure listener status is unchanged, but LB status is changed.
        # LB should still be locked even with pool and subordinate object
        # updates.
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE,
                                            constants.ONLINE)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)

    def test_create_with_listener_id_in_pool_dict(self):
        api_pool = self.create_pool_sans_listener(
            self.lb.get('id'), constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN,
            listener_id=self.listener.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assertEqual(constants.PROTOCOL_HTTP, api_pool.get('protocol'))
        self.assertEqual(constants.LB_ALGORITHM_ROUND_ROBIN,
                         api_pool.get('lb_algorithm'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_with_id(self):
        pid = uuidutils.generate_uuid()
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    id=pid)
        self.assertEqual(pid, api_pool.get('id'))

    def test_create_with_project_id(self):
        pid = uuidutils.generate_uuid()
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    project_id=pid)
        self.assertEqual(self.project_id, api_pool.get('project_id'))

    def test_create_with_duplicate_id(self):
        pool = self.create_pool(self.lb.get('id'),
                                self.listener.get('id'),
                                constants.PROTOCOL_HTTP,
                                constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'), constants.ACTIVE)
        path = self.POOLS_PATH.format(lb_id=self.lb.get('id'),
                                      listener_id=self.listener.get('id'))
        body = {'id': pool.get('id'), 'protocol': constants.PROTOCOL_HTTP,
                'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
                'project_id': self.project_id}
        self.post(path, body, status=409, expect_errors=True)

    def test_bad_create(self):
        api_pool = {'name': 'test1'}
        self.post(self.pools_path, api_pool, status=400)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_with_listener_with_default_pool_id_set(self):
        self.create_pool(self.lb.get('id'),
                         self.listener.get('id'),
                         constants.PROTOCOL_HTTP,
                         constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'), constants.ACTIVE)
        path = self.pools_path_deprecated.format(
            lb_id=self.lb.get('id'), listener_id=self.listener.get('id'))
        body = {'protocol': constants.PROTOCOL_HTTP,
                'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
                'project_id': self.project_id}
        self.post(path, body, status=409, expect_errors=True)

    def test_create_bad_protocol(self):
        pool = {'protocol': 'STUPID_PROTOCOL',
                'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN}
        self.post(self.pools_path, pool, status=400)

    def test_create_with_bad_handler(self):
        self.handler_mock().pool.create.side_effect = Exception()
        self.create_pool(self.lb.get('id'),
                         self.listener.get('id'),
                         constants.PROTOCOL_HTTP,
                         constants.LB_ALGORITHM_ROUND_ROBIN)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ERROR)

    def test_create_over_quota(self):
        self.check_quota_met_true_mock.start()
        self.addCleanup(self.check_quota_met_true_mock.stop)
        body = {'protocol': constants.PROTOCOL_HTTP,
                'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
                'project_id': self.project_id}
        self.post(self.pools_path, body, status=403)

    def test_update(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        new_pool = {'name': 'new_name'}
        self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                 new_pool, status=202)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        response = self.get(self.pool_path.format(pool_id=api_pool.get('id')))
        response_body = response.json
        self.assertNotEqual('new_name', response_body.get('name'))
        self.assertIsNotNone(response_body.get('created_at'))
        self.assertIsNotNone(response_body.get('updated_at'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_bad_update(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'))
        new_pool = {'enabled': 'one'}
        self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                 new_pool, status=400)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_update_with_bad_handler(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        new_pool = {'name': 'new_name'}
        self.handler_mock().pool.update.side_effect = Exception()
        self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                 new_pool, status=202)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ERROR)

    def test_delete(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        api_pool['operating_status'] = constants.ONLINE
        response = self.get(self.pool_path.format(
            pool_id=api_pool.get('id')))
        pool = response.json

        self.assertIsNone(api_pool.pop('updated_at'))
        self.assertIsNotNone(pool.pop('updated_at'))
        self.assertEqual(api_pool, pool)
        self.delete(self.pool_path.format(pool_id=api_pool.get('id')))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_bad_delete(self):
        self.delete(self.pool_path.format(
            pool_id=uuidutils.generate_uuid()), status=404)

    def test_delete_with_l7policy(self):
        api_pool = self.create_pool_sans_listener(
            self.lb.get('id'), constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        self.create_l7policy(self.lb.get('id'), self.listener.get('id'),
                             constants.L7POLICY_ACTION_REDIRECT_TO_POOL,
                             redirect_pool_id=api_pool.get('id'))
        self.set_lb_status(lb_id=self.lb.get('id'))
        self.delete(self.pool_path.format(
            pool_id=api_pool.get('id')), status=409)

    def test_delete_with_bad_handler(self):
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        api_pool['operating_status'] = constants.ONLINE
        response = self.get(self.pool_path.format(
            pool_id=api_pool.get('id')))
        pool = response.json

        self.assertIsNone(api_pool.pop('updated_at'))
        self.assertIsNotNone(pool.pop('updated_at'))
        self.assertEqual(api_pool, pool)
        self.handler_mock().pool.delete.side_effect = Exception()
        self.delete(self.pool_path.format(pool_id=api_pool.get('id')))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ERROR)

    def test_create_with_session_persistence(self):
        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
              "cookie_name": "test_cookie_name"}
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    session_persistence=sp)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        response = self.get(self.pool_path.format(
            pool_id=api_pool.get('id')))
        response_body = response.json
        sess_p = response_body.get('session_persistence')
        self.assertIsNotNone(sess_p)
        self.assertEqual(constants.SESSION_PERSISTENCE_HTTP_COOKIE,
                         sess_p.get('type'))
        self.assertEqual('test_cookie_name', sess_p.get('cookie_name'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_create_with_bad_session_persistence(self):
        sp = {"type": "persistence_type",
              "cookie_name": "test_cookie_name"}
        pool = {'protocol': constants.PROTOCOL_HTTP,
                'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
                'session_persistence': sp}
        self.post(self.pools_path, pool, status=400)

    def test_add_session_persistence(self):
        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
              "cookie_name": "test_cookie_name"}
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(lb_id=self.lb.get('id'))
        response = self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                            body={'session_persistence': sp})
        api_pool = response.json
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.assertNotEqual(sp, api_pool.get('session_persistence'))

    def test_update_session_persistence(self):
        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
              "cookie_name": "test_cookie_name"}
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    session_persistence=sp)
        self.set_lb_status(lb_id=self.lb.get('id'))
        response = self.get(self.pool_path.format(
            pool_id=api_pool.get('id')))
        response_body = response.json
        sess_p = response_body.get('session_persistence')
        sess_p['cookie_name'] = 'new_test_cookie_name'
        api_pool = self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                            body={'session_persistence': sess_p}).json
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.assertNotEqual(sess_p, api_pool.get('session_persistence'))
        self.set_lb_status(self.lb.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_update_preserve_session_persistence(self):
        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
              "cookie_name": "test_cookie_name"}
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    session_persistence=sp)
        self.set_lb_status(lb_id=self.lb.get('id'))
        pool_update = {'lb_algorithm': constants.LB_ALGORITHM_SOURCE_IP}
        api_pool = self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                            body=pool_update).json
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        response = self.get(self.pool_path.format(
            pool_id=api_pool.get('id'))).json
        self.assertEqual(sp, response.get('session_persistence'))

    def test_update_bad_session_persistence(self):
        self.skip('This test should pass after a validation layer.')
        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
              "cookie_name": "test_cookie_name"}
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    session_persistence=sp)
        self.set_lb_status(lb_id=self.lb.get('id'))
        response = self.get(self.pool_path.format(
            pool_id=api_pool.get('id')))
        response_body = response.json
        sess_p = response_body.get('session_persistence')
        sess_p['type'] = 'persistence_type'
        self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                 body={'session_persistence': sess_p}, status=400)

    def test_delete_with_session_persistence(self):
        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
              "cookie_name": "test_cookie_name"}
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    session_persistence=sp)
        self.set_lb_status(lb_id=self.lb.get('id'))
        self.delete(self.pool_path.format(pool_id=api_pool.get('id')),
                    status=202)
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.set_lb_status(self.lb.get('id'))
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.ACTIVE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.ACTIVE, constants.ONLINE)

    def test_delete_session_persistence(self):
        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
              "cookie_name": "test_cookie_name"}
        api_pool = self.create_pool(self.lb.get('id'),
                                    self.listener.get('id'),
                                    constants.PROTOCOL_HTTP,
                                    constants.LB_ALGORITHM_ROUND_ROBIN,
                                    session_persistence=sp)
        self.set_lb_status(lb_id=self.lb.get('id'))
        sp = {'session_persistence': None}
        api_pool = self.put(self.pool_path.format(pool_id=api_pool.get('id')),
                            body=sp, status=202).json
        self.assert_correct_lb_status(self.lb.get('id'),
                                      constants.PENDING_UPDATE,
                                      constants.ONLINE)
        self.assert_correct_listener_status(self.lb.get('id'),
                                            self.listener.get('id'),
                                            constants.PENDING_UPDATE,
                                            constants.ONLINE)
        self.assertIsNotNone(api_pool.get('session_persistence'))

    def test_create_when_lb_pending_update(self):
        self.put(self.LB_PATH.format(lb_id=self.lb.get('id')),
                 body={'name': 'test_name_change'})
        self.post(self.pools_path,
                  body={'protocol': constants.PROTOCOL_HTTP,
                        'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
                        'project_id': self.project_id},
                  status=409)

    def test_update_when_lb_pending_update(self):
        pool = self.create_pool(self.lb.get('id'), self.listener.get('id'),
                                constants.PROTOCOL_HTTP,
                                constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'))
        self.put(self.LB_PATH.format(lb_id=self.lb.get('id')),
                 body={'name': 'test_name_change'})
        self.put(self.pool_path.format(pool_id=pool.get('id')),
                 body={'protocol': constants.PROTOCOL_HTTPS},
                 status=409)

    def test_delete_when_lb_pending_update(self):
        pool = self.create_pool(self.lb.get('id'), self.listener.get('id'),
                                constants.PROTOCOL_HTTP,
                                constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'))
        self.put(self.LB_PATH.format(lb_id=self.lb.get('id')),
                 body={'name': 'test_name_change'})
        self.delete(self.pool_path.format(pool_id=pool.get('id')), status=409)

    def test_create_when_lb_pending_delete(self):
        self.delete(self.LB_DELETE_CASCADE_PATH.format(
            lb_id=self.lb.get('id')))
        self.post(self.pools_path,
                  body={'protocol': constants.PROTOCOL_HTTP,
                        'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
                        'project_id': self.project_id},
                  status=409)

    def test_update_when_lb_pending_delete(self):
        pool = self.create_pool(self.lb.get('id'), self.listener.get('id'),
                                constants.PROTOCOL_HTTP,
                                constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'))
        self.delete(self.LB_DELETE_CASCADE_PATH.format(
            lb_id=self.lb.get('id')))
        self.put(self.pool_path.format(pool_id=pool.get('id')),
                 body={'protocol': constants.PROTOCOL_HTTPS},
                 status=409)

    def test_delete_when_lb_pending_delete(self):
        pool = self.create_pool(self.lb.get('id'), self.listener.get('id'),
                                constants.PROTOCOL_HTTP,
                                constants.LB_ALGORITHM_ROUND_ROBIN)
        self.set_lb_status(self.lb.get('id'))
        self.delete(self.LB_DELETE_CASCADE_PATH.format(
            lb_id=self.lb.get('id')))
        self.delete(self.pool_path.format(pool_id=pool.get('id')), status=409)
